from flask import Blueprint, request, current_app
from flask_jsonpify import jsonpify
from flask.helpers import NotFound

import demjson

from .controllers import Controllers
from .sources import extract_text_fields
from .logger import logger, logging


def default_rules(field):
    if field.get('es:title'):
        if field.get('es:keyword'):
            return ['^10']
        else:
            return ['^3', '.hebrew^10']
    elif field.get('es:boost'):
        return ['^10']
    return ['']


class APIESBlueprint(Blueprint):

    def __init__(self, app,
                 sources,
                 es_client,
                 index_name,
                 document_doctype='document',
                 dont_highlight=[],
                 text_field_rules=default_rules,
                 multi_match_type='most_fields',
                 multi_match_operator='and',
                 debug_queries=False):
        super().__init__('apies', 'apies')
        self.controllers = Controllers(
            index_name=index_name,
            text_fields=extract_text_fields(sources, text_field_rules),
            document_doctype=document_doctype,
            multi_match_type=multi_match_type,
            multi_match_operator=multi_match_operator,
            dont_highlight=dont_highlight,
            debug_queries=debug_queries
        )
        if debug_queries:
            logger.setLevel(logging.DEBUG)
        else:
            logger.setLevel(logging.INFO)
        logger.error('DEBUG')
        logger.debug('DEBUG')

        self.add_url_rule(
            '/get/<path:doc_id>',
            'get_document_handler',
            self.get_document_handler,
            methods=['GET']
        )
        self.add_url_rule(
            '/search/count',
            'simple_count_handler',
            self.count_handler,
            methods=['GET']
        )
        self.add_url_rule(
            '/search/<string:types>',
            'dynamic_search_handler',
            self.search_handler,
            methods=['GET']
        )
        self.add_url_rule(
            '/search/timeline/<string:types>/<string:search_term>/'
            '<string:from_date>/<string:to_date>',
            'timeline_handler',
            self.timeline_handler,
            methods=['GET']
        )

        app.config['ES_CLIENT'] = es_client

    def search_handler(self, types):
        es_client = current_app.config['ES_CLIENT']

        try:
            types_formatted = str(types).split(',')
            filters = request.values.get('filter')
            search_term = request.values.get('q')
            from_date = request.values.get('from_date')
            to_date = request.values.get('to_date')
            size = request.values.get('size', 10)
            offset = request.values.get('offset', 0)
            order = request.values.get('order')
            result = self.controllers.search(
                es_client, types_formatted, search_term,
                from_date, to_date, size, offset, filters,
                score_threshold=0, sort_fields=order
            )
        except Exception as e:
            logger.exception('Error searching %s for types: %s ' % (search_term, str(types)))
            result = {'error': str(e)}
        return jsonpify(result)

    def count_handler(self):
        es_client = current_app.config['ES_CLIENT']

        config = request.values.get('config')
        try:
            config = demjson.decode(config)
            search_term = request.values.get('q')
            from_date = request.values.get('from_date')
            to_date = request.values.get('to_date')
            result = self.controllers.count(
                es_client, search_term, from_date, to_date, config
            )
        except Exception as e:
            logger.exception('Error counting with config %r', config)
            result = {'error': str(e)}
        return jsonpify(result)

    def timeline_handler(self, types, search_term, from_date, to_date):
        es_client = current_app.config['ES_CLIENT']
        index_name = current_app.config['INDEX_NAME']
        text_fields = current_app.config['TEXT_FIELDS']

        try:
            types_formatted = str(types).split(',')
            filters = request.values.get('filter')
            result = self.controllers.timeline(
                es_client, index_name, text_fields,
                types_formatted, search_term,
                from_date, to_date, filters
            )
        except Exception as e:
            logger.exception('Error getting timeline %s for types: %s ' % (search_term, str(types)))
            result = {'error': str(e)}
        return jsonpify(result)

    def get_document_handler(self, doc_id):
        es_client = current_app.config['ES_CLIENT']

        result = self.controllers.get_document(
            es_client, doc_id
        )
        if result is None:
            logger.warning('Failed to fetch document for %r', doc_id)
            raise NotFound()
        return jsonpify(result)



import elasticsearch

from .query import Query


# ### HIGHLIGHT HANDLING
# DONT_HIGHLIGHT = {
#     'kind',
#     'kind_he',
#     'budget_code',
#     'entity_kind',
#     'entity_id',
#     'code',
# }

class Controllers():

    def __init__(self,
                 index_name,
                 text_fields,
                 document_doctype='document',
                 multi_match_type='most_fields',
                 multi_match_operator='and',
                 dont_highlight=tuple(),
                 debug_queries=False):

        self.index_name = index_name
        self.text_fields = text_fields
        self.document_doctype = document_doctype
        self.multi_match_type = multi_match_type
        self.multi_match_operator = multi_match_operator
        self.dont_highlight = dont_highlight
        self.debug_queries = debug_queries

    # REPLACEMENTS
    def _prepare_replacements(self, highlighted):
        return [
            (h.replace('<em>', '').replace('</em>', ''), h)
            for h in highlighted
        ]

    def _do_replacements(self, value, replacements):
        if value is None:
            return None

        if isinstance(value, str):
            for src, dst in replacements:
                value = value.replace(src, dst)
            value = value.replace('</em> <em>', ' ')
            return value

        if isinstance(value, (int, bool, float)):
            return value

        if isinstance(value, list):
            return [self._do_replacements(v, replacements) for v in value]

        if isinstance(value, dict):
            return dict((k, self._do_replacements(v, replacements)) for k, v in value.items())

        assert False, 'Unknown type %r' % value

    def _merge_highlight_into_source(self, source, highlights, dont_highlight):
        for field, highlighted in highlights.items():
            if field in dont_highlight:
                continue
            highlighted = self._prepare_replacements(highlighted)
            field_parts = field.split('.')
            src = source
            field = field_parts[0]
            while len(field_parts) > 1:
                if isinstance(src[field], dict):
                    field_parts.pop(0)
                    src = src[field]
                    field = field_parts[0]
                else:
                    break

            src[field] = self._do_replacements(src[field], highlighted)
        return source

    # UTILS
    def _validate_types(self, text_fields, types):
        if 'all' in types:
            types = text_fields.keys()

        for type_name in types:
            if type_name not in text_fields:
                raise ValueError('not a real type %s' % type_name)
        return types

    # Main API
    def search(self,
               es_client,
               types,
               term,
               from_date,
               to_date,
               size,
               offset,
               filters,
               score_threshold=0,
               sort_fields=None):
        types = self._validate_types(self.text_fields, types)

        query_results = Query(types)
        if term:
            query_results = query_results.apply_term(
                term, self.text_fields,
                multi_match_type=self.multi_match_type,
                multi_match_operator=self.multi_match_operator)

        # Apply the filters
        query_results = query_results.apply_filters(filters)

        # Apply sorting - if there are fields to sort by, apply the scoring as the sorting
        if sort_fields is None:
            if term:
                query_results.apply_scoring()
            else:
                query_results.apply_sorting({'score': {'order': 'desc'}}, 0)
        else:
            query_results.apply_sorting(sort_fields, score_threshold)

        # Apply pagination
        query_results = query_results.apply_pagination(size, offset)

        # Apply highlighting
        if term and self.dont_highlight != '*' and '*' not in self.dont_highlight:
            highlighted = True
            query_results = query_results.apply_highlighting()
        else:
            highlighted = False

        # Apply the time range
        query_results = query_results.apply_time_range(from_date, to_date)\
            .run(es_client, self.index_name, self.debug_queries)

        default_sort_score = (0,)
        if highlighted:
            search_results = [
                dict(
                    source=self._merge_highlight_into_source(
                        hit['_source'],
                        hit['highlight'],
                        self.dont_highlight
                    ),
                    type=hit['_type'],
                    score=hit['_score'] or hit.get('sort', default_sort_score)[0]
                )
                for hit in query_results['hits']['hits']
            ]
        else:
            search_results = [
                dict(
                    source=hit['_source'],
                    type=hit['_type'],
                    score=hit['_score'] or hit.get('sort', default_sort_score)[0]
                )
                for hit in query_results['hits']['hits']
            ]

        return dict(
            search_counts=dict(
                _current=dict(
                    total_overall=query_results['hits']['total']
                )
            ),
            search_results=search_results
        )

    def count(self, es_client, term, from_date, to_date, config):
        counts = {}
        for item in config:
            doc_types = item['doc_types']
            doc_types = self._validate_types(self.text_fields, doc_types)
            filters = item['filters']
            id = item['id']
            query_results = Query(doc_types)
            if term:
                query_results = query_results.apply_term(
                    term, self.text_fields,
                    multi_match_type=self.multi_match_type,
                    multi_match_operator=self.multi_match_operator
                )
            query_results = query_results\
                .apply_filters(filters)\
                .apply_pagination(0, 0)\
                .apply_time_range(from_date, to_date)\
                .run(es_client, self.index_name, self.debug_queries)
            counts[id] = dict(
                total_overall=query_results['hits']['total']
            )
        return dict(
            search_counts=counts
        )

    def timeline(self, es_client, index_name, text_fields,
                 types, term, from_date, to_date, filters):
        types = self._validate_types(text_fields, types)

        query_results = Query(types)\
            .apply_term(
                term, text_fields,
                multi_match_type=self.multi_match_type,
                multi_match_operator=self.multi_match_operator
            )\
            .apply_filters(filters)\
            .apply_pagination(0, 0)\
            .apply_time_range(from_date, to_date)\
            .apply_month_aggregates()\
            .run(es_client, index_name, self.debug_queries)

        timeline = query_results.get('aggregations', {}).get('timeline', {}).get('buckets', [])
        timeline = ((b['key'], b['doc_count'])
                    for b in timeline
                    if len(b['key']) == 7)
        if None not in (from_date, to_date):
            timeline = filter(lambda k: k[0] >= from_date[:7] and k[0] <= to_date[:7],
                            timeline)
        timeline = sorted(timeline)

        return dict(
            timeline=timeline
        )

    def get_document(self, es_client, doc_id):
        try:
            result = es_client.get(self.index_name, doc_id, doc_type=self.document_doctype)
            return result.get('_source')
        except elasticsearch.exceptions.NotFoundError:
            return None



import demjson
import json

from .logger import logger


# ### QUERY DSL HANDLING
class Query():

    def __init__(self, types):
        self.q = {}
        self.types = types

    def __str__(self):
        return demjson.encode(self.q)

    def run(self, es_client, index_name, debug):
        if debug:
            logger.debug('QUERY:\n%s', json.dumps(self.q, indent=2, ensure_ascii=False))
        return es_client.search(
                    index=index_name,
                    doc_type=','.join(self.types),
                    body=self.q
               )

    def must(self):
        return self.q.setdefault('query', {})\
                     .setdefault('function_score', {})\
                     .setdefault('query', {})\
                     .setdefault('bool', {})\
                     .setdefault('must', [])

    def must_not(self):
        return self.q.setdefault('query', {})\
                     .setdefault('function_score', {})\
                     .setdefault('query', {})\
                     .setdefault('bool', {})\
                     .setdefault('must_not', [])

    def apply_term(self, term, text_fields,
                   multi_match_type='most_fields', multi_match_operator='and'):
        search_fields = [text_fields[type_name] for type_name in self.types]
        search_fields = list(set().union(*search_fields))
        search_fields = dict(
            (k, [x[1] for x in search_fields if x[0] == k])
            for k in ('exact', 'inexact', 'natural')
        )
        matchers = []

        # Multimatch
        matchers.append(dict(
            multi_match=dict(
                query=term,
                fields=[f for f in search_fields['inexact']],
                type=multi_match_type,
                operator=multi_match_operator
            )
        ))

        # Common Terms
        for field in search_fields['natural']:
            if '^' in field:
                name, boost = field.split('^')
            else:
                name, boost = field, 1.0
            matchers.append(dict(
                common={
                    name: dict(
                        query=term,
                        boost=float(boost),
                        cutoff_frequency=0.001,
                    )
                }
            ))
        # Tuples
        parts = term.split()
        parts = [term] + parts + [' '.join(z) for z in zip(parts[:-1], parts[1:])]
        for field in search_fields['exact']:
            matchers.append(dict(
                terms={
                    field: tuple(set(parts))
                }
            ))

        # Apply boosters
        if len(matchers) > 0:
            self.must().append(dict(
                bool=dict(
                    should=matchers,
                    minimum_should_match=1
                )
            ))

        return self

    def apply_scoring(self):
        fs = self.q.setdefault('query', {}).setdefault('function_score', {})
        fs.update(dict(
            boost_mode='multiply',
            field_value_factor=dict(
                field='score',
                modifier='sqrt',
                missing=1
            )
        ))
        return self

    def apply_sorting(self, sort_fields, score_threshold):
        # Apply the scoring threshold - since we are no longer sorting by score, it is important to use a score
        # threshold so as not to get irrelevant results
        self.q.setdefault('min_score', score_threshold)
        if isinstance(sort_fields, str):
            if sort_fields[0] == '-':
                sort_fields = sort_fields[1:]
                sort = {sort_fields: {'order': 'desc'}}
            else:
                sort = {sort_fields: {'order': 'asc'}}
        else:
            sort = sort_fields

        # Then sort by the sort fields, for example - {'__last_modified_at': {'order': 'desc'}}
        self.q.setdefault('sort', sort)

        return self

    def apply_pagination(self, page_size, offset):
        self.q.update({
            'size': int(page_size),
            'from': int(offset)
        })
        return self

    def apply_highlighting(self):
        self.q['highlight'] = dict(
            fields={'*': {}}
        )
        return self

    def parse_filter_op(self, k, v):
        must = True
        parts = k.split('__')
        op = None

        if len(parts) > 1 and parts[-1] in ('gt', 'gte', 'lt', 'lte', 'eq', 'not', 'like'):
            op = parts[-1]
            k = '__'.join(parts[:-1])

        if op == 'not':
            must = False
            op = None

        if op is not None:
            if op == 'like':
                ret = dict(
                    match={
                        k: v
                    }
                )
            else:
                ret = dict(
                    range={
                        k: {
                            op: v
                        }
                    }
                )
        else:
            if isinstance(v, list):
                ret = dict(
                    terms={
                        k: v
                    }
                )
            else:
                ret = dict(
                    term={
                        k: v
                    }
                )
        return ret, must

    def apply_filters(self, filters):
        if not filters:
            return self

        if isinstance(filters, str):
            if filters.startswith('[') and filters.endswith(']'):
                pass
            elif not filters.startswith('{'):
                filters = '{' + filters + '}'
            filters = demjson.decode(filters)

        if isinstance(filters, dict):
            filters = [filters]

        should_clauses = []
        if isinstance(filters, list):
            for i in filters:
                bool_clause = {}
                for k, v in i.items():
                    clause, positive = self.parse_filter_op(k, v)
                    if positive:
                        bool_clause.setdefault('must', []).append(clause)
                    else:
                        bool_clause.setdefault('must_not', []).append(clause)
                should_clauses.append(dict(bool=bool_clause))
        self.must().append(dict(bool=dict(should=should_clauses)))
        return self

    def apply_time_range(self, from_date, to_date):
        if None not in (from_date, to_date):
            self.must().extend([
                dict(
                    range=dict(
                        __date_range_from=dict(
                            lte=to_date
                        )
                    )
                ),
                dict(
                    range=dict(
                        __date_range_to=dict(
                            gte=from_date
                        )
                    )
                ),
            ])
        return self

    def apply_month_aggregates(self):
        self.q.setdefault('aggs', {}).update(dict(
            timeline=dict(
                terms=dict(
                    field='__date_range_months',
                    size=2500,
                    order=dict(_term='asc')
                )
            )
        ))
        return self
